using Pkg

using CountTimeSeries, Plots, CSV, Shapefile, Images, ImageIO
using PlotShapefiles, Geodesy, StringDistances
using LinearAlgebra, Optim, Dates, Distributions, Random
using StatsPlots, JLD2, DataFrames

cd("C:\\Users\\stapperm\\sciebo\\Arbeit\\Projects\\CountTimeSeries_Review")

# Settings for plots
gr(tickfontsize = 9, guidefontsize = 12, linewidth = 2)

# Read in Germany Shapefile
Ger = Shapefile.Table(".\\Application 1\\Data\\LKGermany.shp")
GerGeom = Shapefile.shapes(Ger)

# Reading in COVID Data
# dat = CSV.read(".\\Application 1\\Data\\RKI_COVID19.csv")
dat = CSV.read(".\\Application 1\\Data\\RKI_COVID19_update.csv", DataFrame)

# Transform
IDs = sort(unique(dat.IdLandkreis))
dates = sort(unique(dat.Meldedatum))
dates[1]
dates[end]
dates2 = collect(Date(2020,1,7):Day(1):Date(2021,3,15))
cases = fill(0, (length(IDs), length(dates2)))

# Translating dates
function str2date(x)
    split1 = split(x, "/")
    y = parse(Int64, split1[1])
    m = parse(Int64, split1[2])
    d = parse(Int64, split(split1[3], " ")[1])
    Date(y, m, d)
end
dat.Meldedatum = str2date.(dat.Meldedatum)
dat.AnzahlFall[dat.AnzahlFall .< 0] .= 0

for i = 1:size(dat)[1]
    IDLK = findall(dat.IdLandkreis[i] .== IDs)[1]
    IDdate = findall(dat.Meldedatum[i] .== dates2)[1]
    cases[IDLK, IDdate] += dat.AnzahlFall[i]
end

# Merging Berlin
# In number of cases
cases[325, :] = sum(cases[325:336, :], dims = 1)[1, :]
cases[326:401, :] = cases[337:412, :]
cases = cases[1:401, :]

# and the LK names
LKnames = unique(dat.Landkreis)
LKnames[325] = "SK Berlin"
LKnames[326:401] = LKnames[337:412]
LKnames = LKnames[1:401]

# Rename Shapefiles, no doubles
function checkDouble(x)
    ind = fill(false, length(x))
    for i = 1:length(x)
        if sum(x[i] .== x) > 1
            ind[i] = true
        end
    end
    unique(x[ind])
end

doubleNames = checkDouble(Ger.name_2)
LKfirst = fill(true, length(doubleNames))

LKfirst[[1, 5, 10, 11, 12, 13, 17, 20, 22, 23]] .= false
GerNames = copy(Ger.name_2)

for i = 1:length(doubleNames)
    ind = findall(Ger.name_2 .== doubleNames[i])
    if LKfirst[i]
        global GerNames[ind[1]] = "LK "*Ger.name_2[ind[1]]
        global GerNames[ind[2]] = "SK "*Ger.name_2[ind[2]]
    else
        global GerNames[ind[1]] = "SK "*Ger.name_2[ind[1]]
        global GerNames[ind[2]] = "LK "*Ger.name_2[ind[2]]
    end
end

# Cases in Limburg and neighbouring districts
plot(cases[findall(LKnames .== "LK Limburg-Weilburg")[1], :])

findall(LKnames .== "LK Lahn-Dill-Kreis")[1]
findall(LKnames .== "LK Hochtaunuskreis")[1]
findall(LKnames .== "LK Rheingau-Taunus-Kreis")[1]
findall(LKnames .== "LK Rhein-Lahn-Kreis")[1]
findall(LKnames .== "LK Westerwaldkreis")[1]

yLW = cases[findall(LKnames .== "LK Limburg-Weilburg")[1], :]
NGB = cases[[132, 124, 129, 152, 153], :]


mean(yLW .== 0)
mean(yLW .> 50)

mean(yLW)
var(yLW)

# Rolling mean for tidier plot
function rollmean(x, m)
    T = length(x)
    out = zeros(Union{Missing, Float64}, T)
    out[1:m] .= missing
    out[T-m+1:T] .= missing
    for t = m+1:T-m
        out[t] = mean(x[t-m:t+m])
    end
    out
end

# Indices of first days of months
indfirst = 26 .+ cumsum([0, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31, 28])

plot(dates2, yLW, xlabel = "Time", ylabel = "Daily New Infections",
     title = "", label = "")
xticks!(datetime2rata.(dates2[indfirst[1:2:14]]), ["Feb '20", "Apr '20", "Jun '20", "Aug '20", "Oct '20", "Dec '20", "Feb '21"])
savefig(".\\Application 1\\Plots\\LWTotal.pdf")

findall(LKnames .== "LK Lahn-Dill-Kreis")[1]
findall(LKnames .== "LK Hochtaunuskreis")[1]
findall(LKnames .== "LK Rheingau-Taunus-Kreis")[1]
findall(LKnames .== "LK Rhein-Lahn-Kreis")[1]
findall(LKnames .== "LK Westerwaldkreis")[1]

plot(dates2, rollmean(yLW, 3), label = "Limburg-Weilburg", legend = :topleft,
    xlabel = "Time", ylabel = "Daily New Infections (Rolling Mean)")
xticks!(datetime2rata.(dates2[indfirst[1:2:14]]), ["Feb '20", "Apr '20", "Jun '20", "Aug '20", "Oct '20", "Dec '20", "Feb '21"])
plot!(dates2, rollmean(NGB[1, :], 3), label = "Lahn-Dill-Kreis", linealpha = 0.5)
plot!(dates2, rollmean(NGB[2, :], 3), label = "Hochtaunuskreis", linealpha = 0.5)
plot!(dates2, rollmean(NGB[3, :], 3), label = "Rheingau-Taunus-Kreis", linealpha = 0.5)
plot!(dates2, rollmean(NGB[4, :], 3), label = "Rhein-Lahn-Kreis", linealpha = 0.5)
plot!(dates2, rollmean(NGB[5, :], 3), label = "Westerwaldkreis", linealpha = 0.5)
plot!(dates2, rollmean(yMS, 3), label = "", color = 1)

savefig(".\\Application 1\\Plots\\RollMean.pdf")

acf = (i -> cor(yLW[1:end-i], yLW[i+1:end])).(1:21)

bar(1:21, acf, bar_width = 0.2, label = "", linewidth = 0)
ylims!((-0.01, 1))
xticks!(1:21, string.(1:21))
xlabel!("Lag")
ylabel!("Correlation")
savefig(".\\Application 1\\Plots\\AcfLW.pdf")

y = yLW[8:end]
dates2 = dates2[8:end]

X = log.(NGB[:, 1:end-7] .+ 1.0)

# Try models on full time series
modelP = Model(pastObs = [1, 7], pastMean = 1, X = X, external = fill(false, 5),
               link = "Log", distr = "Poisson")
modelNB = Model(pastObs = [1, 7], pastMean = 1, X = X, external = fill(false, 5),
                link = "Log", distr = "NegativeBinomial")
settingP = MLESettings(y, modelP, [1; fill(0.2, 3); fill(0.1, 5)], optimizer = "NelderMead", ci = true)
resP = fit(y, modelP, settingP)
plot(y)
plot!(resP.λ)
resQP = QPois(resP)
settingNB = MLESettings(y, modelNB, resQP.θ, optimizer = "NelderMead", ci = true)
resNB = fit(y, modelNB, settingNB)

sum(resP.pars.α) + sum(resP.pars.β)
sum(resNB.pars.α) + sum(resNB.pars.β)

# Successive estimation and prediction
# Use first 322, 323, 324, ..., 420 observations for estimation
# Then predict the next 7 days
results = Array{INGARCHresults, 2}(undef, 99, 3)
PRED = zeros(99, 7, 3)
PREDci = zeros(99, 2, 7, 3)

Random.seed!(05052021)
for i = 1:99
    # Specify Poisson Model
    modelP = Model(pastObs = [1, 7], pastMean = 1, X = X[:, 1:321+i], external = fill(false, 5),
                   link = "Log", distr = "Poisson")
    # Specify NegativeBinomial Model
    modelNB = Model(pastObs = [1, 7], pastMean = 1, X = X[:, 1:321+i], external = fill(false, 5),
                    link = "Log", distr = "NegativeBinomial")
    # Settings:
    settingP = MLESettings(y[1:321+i], modelP, resP.θ, optimizer = "NelderMead", ci = false)
    settingNB = MLESettings(y[1:321+i], modelNB, resNB.θ, optimizer = "NelderMead", ci = false)

    results[i, 1] = fit(y[1:321+i], modelP, settingP, printResults = false)
    results[i, 2] = QPois(results[i, 1])
    results[i, 3] = fit(y[1:321+i], modelNB, settingNB, printResults = false)

    # Build new regressor matrix for forecast
    Xnew = X[:, 322+i:328+i] .+ 0.0

    pred = predict(results[i, 1], 7, 10000, Xnew)
    PRED[i, :, 1] = pred[1]
    PREDci[i, :, :, 1] = pred[2]
    pred = predict(results[i, 2], 7, 10000, Xnew)
    PRED[i, :, 2] = pred[1]
    PREDci[i, :, :, 2] = pred[2]
    pred = predict(results[i, 3], 7, 10000, Xnew)
    PRED[i, :, 3] = pred[1]
    PREDci[i, :, :, 3] = pred[2]

    println(i)
end

# Plotting results:
h = 1
plot(322+h:420+h, y[322+h:420+h], color = 1, xlabel = "Time",
     ylabel = "Daily New Infections", label = "Observed")
plot!(322+h:420+h, PRED[:, h, 1], color = 2, label = "Poisson")
plot!(322+h:420+h, PRED[:, h, 2], color = 3, label = "Quasi Poisson")
plot!(322+h:420+h, PRED[:, h, 3], color = 4, label = "Negative Binomial")
plot!(322+h:420+h, y[322+h:420+h], color = 1, label = "")
xticks!([323, 354, 385, 413], ["Dec '20", "Jan '21", "Feb '21", "Mar '21"])

savefig(".\\Application 1\\Plots\\1ahead.pdf")

h = 7
plot(322+h:420+h, y[322+h:420+h], color = 1, xlabel = "Time",
     ylabel = "Daily New Infections", label = "Observed")
plot!(322+h:420+h, PRED[:, h, 1], color = 2, label = "Poisson")
plot!(322+h:420+h, PRED[:, h, 2], color = 3, label = "Quasi Poisson")
plot!(322+h:420+h, PRED[:, h, 3], color = 4, label = "Negative Binomial")
plot!(322+h:420+h, y[322+h:420+h], color = 1, label = "")
xticks!([354, 385, 413], ["Jan '21", "Feb '21", "Mar '21"])

savefig(".\\Application 1\\Plots\\7ahead.pdf")

h = 7
mean(PREDci[:, 1, h, 1] .<= y[322+h:420+h] .<= PREDci[:, 2, h, 1])
mean(PREDci[:, 1, h, 2] .<= y[322+h:420+h] .<= PREDci[:, 2, h, 2])
mean(PREDci[:, 1, h, 3] .<= y[322+h:420+h] .<= PREDci[:, 2, h, 3])

h = 7
sqrt(mean((PRED[:, h, 1] .- y[322+h:420+h]).^2))
sqrt(mean((PRED[:, h, 2] .- y[322+h:420+h]).^2))
sqrt(mean((PRED[:, h, 3] .- y[322+h:420+h]).^2))

mean(abs.(PRED[:, h, 1] .- y[322+h:420+h]))
mean(abs.(PRED[:, h, 2] .- y[322+h:420+h]))
mean(abs.(PRED[:, h, 3] .- y[322+h:420+h]))

median(abs.(PRED[:, h, 1] .- y[322+h:420+h]))
median(abs.(PRED[:, h, 2] .- y[322+h:420+h]))
median(abs.(PRED[:, h, 3] .- y[322+h:420+h]))

plot((h -> sqrt(mean((PRED[:, h, 1] .- y[322+h:420+h]).^2))).(1:7), xlab = "Horizon",
     ylab = "RMSPE", label = "Poisson", legend = :topleft)
plot!((h -> sqrt(mean((PRED[:, h, 2] .- y[322+h:420+h]).^2))).(1:7),
      label = "Quasi Poisson")
plot!((h -> sqrt(mean((PRED[:, h, 3] .- y[322+h:420+h]).^2))).(1:7),
      label = "NegativeBinomial")
#
plot((h -> mean(abs.(PRED[:, h, 1] .- y[322+h:420+h]))).(1:7), xlab = "Horizon",
     ylab = "MAPE", label = "Poisson", legend = :topleft)
plot!((h -> mean(abs.(PRED[:, h, 2] .- y[322+h:420+h]))).(1:7),
      label = "Quasi Poisson")
plot!((h -> mean(abs.(PRED[:, h, 3] .- y[322+h:420+h]))).(1:7),
      label = "NegativeBinomial")

#
plot((h -> median(abs.(PRED[:, h, 1] .- y[322+h:420+h]))).(1:7), xlab = "Horizon",
     ylab = "MedAPE", label = "Poisson", legend = :topleft)
plot!((h -> median(abs.(PRED[:, h, 2] .- y[322+h:420+h]))).(1:7),
      label = "Quasi Poisson")
plot!((h -> median(abs.(PRED[:, h, 3] .- y[322+h:420+h]))).(1:7),
      label = "NegativeBinomial")

# Better to summarize in table
