using Pkg
# Adding CountTimeSeries package in Pkg REPL
# add https://github.com/ManuelStapper/CountTimeSeries.jl
using CountTimeSeries, Plots, CSV, Shapefile, Images, ImageIO
using PlotShapefiles, Geodesy, StringDistances
using LinearAlgebra, Optim, Dates, Distributions, Random
using StatsPlots, JLD2, DataFrames

cd("C:\\Users\\stapperm\\sciebo\\Arbeit\\Projects\\CountTimeSeries_Review")

gr(tickfontsize = 9, guidefontsize = 12, linewidth = 2)

dat = CSV.read(".\\Application 2\\Data\\animal.csv", DataFrame)

plot(dat.Anorexia, label = "Anorexia", xlabel = "Time", ylabel = "Count",
     legend = :top)
plot!(dat.Lesions, label = "Lesions")
xticks!(1:12:84, string.(2003:2009))
yticks!(0:2:10, string.(0:2:10))

savefig(".\\Application 2\\Plots\\ts.pdf")

function myacf(x, lagMax)
    out = ones(lagMax)

    for i = 1:lagMax
        out[i] = cor(x[i+1:end], x[1:end-i])
    end

    out
end

acfA = myacf(dat.Anorexia, 12)
acfL = myacf(dat.Lesions, 12)

plot([1, 1] .- 0.05, [0, acfA[1]], label = "Anorexia", color = 1)
plot!([1, 1] .+ 0.05, [0, acfL[1]], label = "Lesions", color = 2)
for i = 2:length(acfA)
    plot!([i, i] .- 0.05, [0, acfA[i]], label = "", color = 1)
    plot!([i, i] .+ 0.05, [0, acfL[i]], label = "", color = 2)
end
hline!([0], color = "black", label = "")
ylabel!("ACF")
xlabel!("Lag")
xticks!(1:12, string.(1:12))
yticks!(-0.2:0.2:1, string.(-0.2:0.2:1))

savefig(".\\Application 2\\Plots\\acf.pdf")

models = Array{INARModel, 1}(undef, 8)
models[1] = Model(model = "INARMA", pastObs = 1,
                       zi = true, distr = "Poisson")
models[2] = Model(model = "INARMA", pastObs = 1:2,
                       zi = true, distr = "Poisson")
models[3] = Model(model = "INARMA", pastObs = 1,
                       zi = true, distr = "NegativeBinomial")
models[4] = Model(model = "INARMA", pastObs = 1:2,
                       zi = true, distr = "NegativeBinomial")
models[5] = Model(model = "INARMA", pastObs = 1,
                       distr = "Poisson")
models[6] = Model(model = "INARMA", pastObs = 1:2,
                       distr = "Poisson")
models[7] = Model(model = "INARMA", pastObs = 1,
                       distr = "NegativeBinomial")
models[8] = Model(model = "INARMA", pastObs = 1:2,
                       distr = "NegativeBinomial")

settingAno = MLESettings.(fill(dat.Anorexia, 8), models, optimizer = "NelderMead", ci = true)
settingLes = MLESettings.(fill(dat.Lesions, 8), models, optimizer = "NelderMead", ci = true)

results = Array{INARMAresults, 2}(undef, (8, 2))
results[:, 1] = fit.(fill(dat.Anorexia, 8), models, settingAno, printResults = false)
results[:, 2] = fit.(fill(dat.Lesions, 8), models, settingLes, printResults = false)

AIC.(results, 2)
