using Pkg
# Adding CountTimeSeries package in Pkg REPL
# add https://github.com/ManuelStapper/CountTimeSeries.jl
using CountTimeSeries, Plots, CSV, Shapefile, Images, ImageIO
using PlotShapefiles, Geodesy, StringDistances
using LinearAlgebra, Optim, Dates, Distributions, Random
using StatsPlots, JLD2, DataFrames, Luxor

gr(tickfontsize = 9, guidefontsize = 12, linewidth = 2)
cd("C:\\Users\\stapperm\\sciebo\\Arbeit\\Projects\\CountTimeSeries_Review")

# include(".\\Application 3\\MapProjections.jl")

# Read in data
datRaw = readlines(".\\Application 3\\Data\\insolvencies.txt")
datRaw2 = fill(0, (42, 108))

# Indices excluding year totals
ind = setdiff(collect(1:117), collect(0:13:117))

# Tranforming string to numbers
function f(x)
    if x == "-"
        return 0
    end
    temp = split(x, ".")
    sum(parse.(Int64, temp).*1000 .^((length(temp) - 1):-1:0))
end

# Filling data matrix
for i = 1:42
    raw1 = split(datRaw[i], "\t")
    datRaw2[i, :] = f.(raw1[ind .+ 1])
end

# Set up district names
names = ["Ahrweiler", "Altenkirchen", "Alzey-Worms", "Bad Dürkheim",
    "Bad Kreuznach", "Bernkastel-Wittlich", "Birkenfeld", "Cochem-Zell",
    "Donnersbergkreis", "Eifelkreis Bitburg-Prüm", "Frankenthal, kfr. S.",
    "Germersheim", "Kaiserslautern, kfr. S.", "Kaiserslautern",
    "Koblenz, kfr. S.", "Kusel", "Landau i.d.P., kfr. S.",
    "Ludwigshafen, kfr. S.", "Mainz, kfr. S.", "Mainz-Bingen",
    "Mayen-Koblenz", "Neustadt a.d.W., kfr. S.",
    "Neuwied", "Pirmasens, kfr. S.", "Rhein-Hunsrück-Kreis",
    "Rhein-Lahn-Kreis", "Rhein-Pfalz-Kreis",
    "Speyer, kfr. S.", "Südliche Weinstraße", "Südwestpfalz",
    "Trier, kfr. S.", "Trier-Saarburg", "Vulkaneifel",
    "Westerwaldkreis", "Worms, kfr. S.", "Zweibrücken, kfr. S."]

# Filling data matrix in correct order
dat = datRaw2[[4,5,31,32,6,16,7,8,33,17,21,34,22,35,3,36,23,24,25,39,
               9,26,10,27,11,12,38,28,37,40,15,19,18,13,29,30], :]

# Descriptives
datTotal = sum(dat, dims = 1)[1, :]

plot(datTotal, label = "")
xticks!(1:12:108, string.(2008:2016))
xlabel!("Time")
ylabel!("Count")

savefig(".\\Application 3\\Plots\\TS.pdf")


# Setting up models
models = Array{Any, 1}(undef, 54)
counter = 1
xs = [[], Array{Float64, 2}(collect(1:108)'), [fill(0.0, 24); Float64.(25:108)]]

for x = 1:3
    global counter
    for pois = [true, false]
        models[counter] = Model(model = "INGARCH",
                            distr = ifelse(pois, "Poisson", "NegativeBinomial"),
                            link = "Log",
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INGARCH",
                            distr = ifelse(pois, "Poisson", "NegativeBinomial"),
                            pastObs = 1,
                            link = "Log",
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INGARCH",
                            distr = ifelse(pois, "Poisson", "NegativeBinomial"),
                            pastObs = 1:2,
                            link = "Log",
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INGARCH",
                            distr = ifelse(pois, "Poisson", "NegativeBinomial"),
                            pastObs = 1,
                            pastMean = 1,
                            link = "Log",
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INGARCH",
                            distr = ifelse(pois, "Poisson", "NegativeBinomial"),
                            pastObs = 1:2,
                            pastMean = 1,
                            link = "Log",
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INARMA",
                            distr = fill(ifelse(pois, "Poisson", "NegativeBinomial"), 2),
                            pastObs = 1,
                            link = "Log",
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INARMA",
                            distr = fill(ifelse(pois, "Poisson", "NegativeBinomial"), 2),
                            pastObs = 1:2,
                            link = fill("Log", 2),
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INARMA",
                            distr = fill(ifelse(pois, "Poisson", "NegativeBinomial"), 2),
                            pastObs = 1,
                            pastMean = 1,
                            link = fill("Log", 2),
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
        models[counter] = Model(model = "INARMA",
                            distr = fill(ifelse(pois, "Poisson", "NegativeBinomial"), 2),
                            pastObs = 1:2,
                            pastMean = 1,
                            link = fill("Log", 2),
                            X = xs[x],
                            external = ifelse(x == 1, [], true))
        counter += 1
    end
end


# Preparing result/initial value/settings arrays
results = Array{Any, 2}(undef, (54, 36))
inits = Array{parameter, 2}(undef, (54, 36))
settings = Array{MLEControl, 2}(undef, (54, 36))

for i = 1:36
    y = dat[i, :]

    for m = 1:54
        nϕ = sum(models[m].distr .== "NegativeBinomial")

        if nϕ == 2
            if length(models[m].external) == 0
                nϕ = 1
            else
                if sum(models[m].external) == 0
                    nϕ = 1
                end
            end
        end

        if typeof(models[m]) != IIDModel
            p = length(models[m].pastObs)
        else
            p = 0
        end

        if typeof(models[m]) in [INGARCHModel, INARMAModel, INMAModel]
            q = length(models[m].pastMean)
        else
            q = 0
        end

        r = size(models[m].X)[1]
        loglin = models[m].link[1] == "Log"

        θ = [ifelse(loglin, log(mean(y)), mean(y));
             fill(0.1, p+q);
             fill(0, r);
             fill(10, nϕ)]
        θ = convert(Array{Float64, 1}, θ)
        inits[m, i] = θ2par(θ, models[m])

        while !isfinite(ll(dat[i, :], models[m], inits[m, i])[1])
            inits[m, i] = θ2par(θ .+ rand(Normal(), length(θ)), models[m])
        end

        # NelderMead not allowed for univariate optimization
        if (length(θ) == 1)
            optimizer = "BFGS"
        else
            optimizer = "NelderMead"
        end

        settings[m, i] = MLEControl(inits[m, i], optimizer, false, Int64(1e3))
    end
end

# Problems with estimation in case of no marginal overdispersion

# Estimation
for m = 1:54, i = 1:36
    println([m, i])
    Random.seed!(m + i*100)
    empUD = mean(dat[i, :]) > var(dat[i, :])
    if !(any(models[m].distr .== "NegativeBinomial") & empUD)
        results[m, i] = fit(dat[i, :], models[m], settings[m, i], printResults = false)
    end
end

###############
### Results ###
###############

# Prepare maps
# Read in shapefile
Ger = Shapefile.Table(".\\Application 3\\Data\\LKGermany.shp")
GerGeom = Shapefile.shapes(Ger)

nutsShp = Shapefile.Table(".\\Application 3\\Data\\nuts2.shp")
iDE = findall(nutsShp.CNTR_CODE .== "DE")
RPgeom2 = Shapefile.shapes(nutsShp)[iDE[[14, 21, 34]]]

# Create Data Frame
iRP = findall(Ger.name_1 .== "Rheinland-Pfalz")
id = [85, 357, 151, 399, 242, 323, 212, 243, 86, 200, 310, 333, 213, 11, 116,
      269, 38, 300, 377, 358, 182, 214, 270, 215, 68, 324, 183, 349, 39, 244,
      184, 131, 283, 350, 12, 257]

nuts2 = [1, 1, 3, 3, 1, 2, 1, 1, 3, 2, 3, 3, 3, 3, 1, 3, 3, 3,
         3, 3, 1, 3, 1, 3, 1, 1, 3, 3, 3, 3, 2, 2, 2, 1, 3, 3]

DAT = DataFrame(Name = names,
                nuts2 = nuts2,
                Counts = Array{Array{Int64, 1}, 1}(undef, 36),
                Polygon = GerGeom[id])

for i = 1:36
    DAT[i, 3] = dat[i, :]
end

# Color Scale: dark red, light red, gray, light green, dark green
colScale = RGB.([127, 255, 186, 137, 10]./255,
                [10, 127, 186, 255, 127]./255,
                [0, 117, 186, 127, 0]./255)

# Find out best models
aics = zeros(54, 36)
for m = 1:54
    for i = 1:36
        empUD = mean(dat[i, :]) > var(dat[i, :])
        if !(any(models[m].distr .== "NegativeBinomial") & empUD)
            aics[m, i] = AIC(results[m, i], 2)
        else
            aics[m, i] = Inf
        end
    end
end

mc = fill(0, 36)
for i = 1:36
    mc[i] = findall(aics[:, i] .== minimum(aics[:, i]))[1]
end

# Summarizing models
idINGARCH = (typeof.(models[mc]) .<: INGARCH) .& (typeof.(models[mc]) != IIDModel)
idIID = typeof.(models[mc]) .== IIDModel

# no INARMA model chosen

idPois = (x -> x.distr == "Poisson").(models[mc])
orderPO = zeros(Int64, 36)
orderPM = zeros(Int64, 36)
Xind = fill(0, 36)
# No trend / starting 2010 / starting directly
trend = fill(0, 36)

for i = 1:36
    if typeof(models[mc[i]]) != IIDModel
        orderPO[i] = length(models[mc[i]].pastObs)
    end
    if typeof(models[mc[i]]) == INGARCHModel
        orderPM[i] = length(models[mc[i]].pastMean)
    end

    XX = results[mc[i], i].model.X
    if size(XX)[1] == 0
        trend[i] = 3
        Xind[i] = 0
    else
        ηTemp = results[mc[i], i].pars.η[1]
        if XX[1] == 0
            Xind[i] = 1
            if ηTemp < 0
                trend[i] = 4
            else
                trend[i] = 2
            end
        else
            Xind[i] = 2
            if ηTemp < 0
                trend[i] = 5
            else
                trend[i] = 1
            end
        end
    end
end

function temp(i)
    p1 = ifelse(idPois[i], "Pois-", "NB-")
    p2 = ifelse(idIID[i], "IID", ifelse(orderPM[i] == 0, "INARCH", "INGARCH"))
    p3 = ifelse(orderPO[i] .> 0, "("*string(orderPO[i]), "")
    p4 = ifelse(orderPM[i] > 0, " ,"*string(orderPM[i])*")", ifelse(idIID[i], "", ")"))
    p5 = [" No", " Partly", " Yes"][Xind[i] + 1]

    return p1*p2*p3*p4*p5
end

# Plotting
center = zeros(2)
t = maximum((x -> x.MBR.top).(DAT[:, 4]))
b = minimum((x -> x.MBR.bottom).(DAT[:, 4]))
l = minimum((x -> x.MBR.left).(DAT[:, 4]))
r = maximum((x -> x.MBR.right).(DAT[:, 4]))
center[2] = mean([t, b])
center[1] = mean([l, r])

gr(size = (900, 1200))
Plots.plot(proj(center, DAT[1, 4]), color = colScale[trend[1]],
           framestyle = :box,
           aspect_ratio = 1)
#
xaxis!([], [])
yaxis!([], [])
for i = 36:-1:1
    Plots.plot!(proj(center, DAT[i, 4]), color = colScale[trend[i]])
end

Plots.plot!(proj(center, RPgeom2[1]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[2]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[3]), color = :transparent, linewidth = 4)

yseq = collect(48.75:0.25:51)
xseq = collect(6:0.25:8.75)

for i = yseq
    temp2 = proj(center, hcat(xseq, fill(i, length(xseq))))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

for i = xseq
    temp2 = proj(center, hcat(fill(i, length(yseq)), yseq))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

Plots.plot!(Shape([-0.03, -0.018, -0.018, -0.03, -0.03].+0.0155,  [0.007, 0.007, 0.017, 0.017, 0.007].-0.0249),
            color = "white")
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155,  [0.014, 0.014, 0.016, 0.016, 0.014].-0.0249),
            color = colScale[3])
#
Plots.plot!(Shape([-0.029, -0.027, -0.027].+0.0155, [0.011, 0.011, 0.013].-0.0249), color = colScale[4], linewidth = 0)
Plots.plot!(Shape([-0.029, -0.029, -0.027].+0.0155, [0.011, 0.013, 0.013].-0.0249), color = colScale[2], linewidth = 0)

Plots.plot!(Shape([-0.029, -0.027, -0.027].+0.0155, [0.008, 0.008, 0.01].-0.0249), color = colScale[5], linewidth = 0)
Plots.plot!(Shape([-0.029, -0.029, -0.027].+0.0155, [0.008, 0.01, 0.01].-0.0249), color = colScale[1], linewidth = 0)

Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029].+0.0155, [0.011, 0.011, 0.013, 0.013].-0.0249), color = :transparent)
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029].+0.0155, [0.008, 0.008, 0.01, 0.01].-0.0249), color = :transparent)

Plots.annotate!([(-0.026+0.0155, 0.015-0.0249, Plots.text("No Trend", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.012-0.0249, Plots.text("After 2009", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.009-0.0249, Plots.text("Complete", :left, 24))])

savefig(".\\Application 3\\Plots\\TrendType.pdf")


gr(size = (900, 1200))
Plots.plot(proj(center, DAT[1, 4]), color = idINGARCH[1] + 1, framestyle = :box)
for i = 36:-1:2
    Plots.plot!(proj(center, DAT[i, 4]), color = idINGARCH[i] + 1)
end

Plots.plot!(proj(center, RPgeom2[1]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[2]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[3]), color = :transparent, linewidth = 4)

for i = yseq
    temp2 = proj(center, hcat(xseq, fill(i, length(xseq))))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

for i = xseq
    temp2 = proj(center, hcat(fill(i, length(yseq)), yseq))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

xaxis!([], [])
yaxis!([], [])
# xlims!((5.8, 8.8))
# ylims!((48.9, 51.0))
Plots.plot!(Shape([-0.03, -0.018, -0.018, -0.03, -0.03].+0.0155,  [0.01, 0.01, 0.017, 0.017, 0.01].-0.0249),
            color = "white")
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155,  [0.014, 0.014, 0.016, 0.016, 0.014].-0.0249), color = 1)
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029].+0.0155, [0.011, 0.011, 0.013, 0.013].-0.0249), color = 2)
Plots.annotate!([(-0.026+0.0155, 0.015-0.0249, Plots.text("IID", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.012-0.0249, Plots.text("INGARCH", :left, 24))])

savefig(".\\Application 3\\Plots\\ModelType.pdf")



gr(size = (900, 1200))
Plots.plot(proj(center, DAT[1, 4]), color = idPois[1] + 1, framestyle = :box)
for i = 36:-1:2
    Plots.plot!(proj(center, DAT[i, 4]), color = idPois[i] + 1)
end
Plots.plot!(proj(center, RPgeom2[1]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[2]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[3]), color = :transparent, linewidth = 4)

yseq = collect(48.75:0.25:51)
xseq = collect(6:0.25:8.75)

for i = yseq
    temp2 = proj(center, hcat(xseq, fill(i, length(xseq))))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

for i = xseq
    temp2 = proj(center, hcat(fill(i, length(yseq)), yseq))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end
xaxis!([], [])
yaxis!([], [])

Plots.plot!(Shape([-0.03, -0.018, -0.018, -0.03, -0.03].+0.0155,  [0.008, 0.008, 0.017, 0.017, 0.008].-0.0249),
            color = "white")
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155, [0.014, 0.014, 0.016, 0.016, 0.014].-0.0249), color = 1)
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029].+0.0155, [0.009, 0.009, 0.011, 0.011].-0.0249), color = 2)
Plots.annotate!([(-0.026+0.0155, 0.015-0.0249, Plots.text("Negative\nBinomial", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.01-0.0249, Plots.text("Poisson", :left, 24))])

savefig(".\\Application 3\\Plots\\Distr.pdf")


function convertPoint(x::Shapefile.Point)
    Luxor.Point(x.x, x.y)
end

xseq = collect(5.8:0.03:8.8) .+ 0.005
yseq = collect(48.9:0.03:51) .+ 0.005
xseq2 = repeat(xseq, inner = length(yseq))
yseq2 = repeat(yseq, outer = length(xseq))
scat = Luxor.Point.(xseq2, yseq2)
FLAG = fill(false, length(scat))

for i = 1:length(scat)
    indic = try
        isinside(scat[i],
                Luxor.Point.([5.85, 5.85, 5.85, 5.85], [50.32, 50.32, 50.41, 50.41]),
                allowonedge = false)
    catch
        false
    end

    if indic
        FLAG[i] = true
    end

    for j = 1:3
        indic = try
            isinside(scat[i], convertPoint.(RPgeom2[j].points), allowonedge = false)
        catch
            false
        end

        if indic
            FLAG[i] = true
        end
    end
end

gr(size = (900, 1200))
Plots.plot(framestyle = :box)
for i = reverse(findall(orderPM .== 1))
    Plots.plot!(proj(center, DAT[i, 4]), color = colScale[orderPO[i] + 3])
end

scat2 = proj(center, hcat((x -> x.x).(scat[FLAG]), (x -> x.y).(scat[FLAG])))
StatsPlots.scatter!(scat2[:, 1], scat2[:, 2], markerstrokewidth = 0,
        color = "black", markersize = 3)
#


xaxis!([], [])
yaxis!([], [])
# xlims!((5.8, 8.8))
# ylims!((48.9, 51.0))

for i = reverse(findall(orderPM .== 0))
    Plots.plot!(proj(center, DAT[i, 4]), color = colScale[orderPO[i] + 3])
end
Plots.plot!(proj(center, RPgeom2[1]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[2]), color = :transparent, linewidth = 4)
Plots.plot!(proj(center, RPgeom2[3]), color = :transparent, linewidth = 4)

yseq = collect(48.75:0.25:51)
xseq = collect(6:0.25:8.75)

for i = yseq
    temp2 = proj(center, hcat(xseq, fill(i, length(xseq))))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

for i = xseq
    temp2 = proj(center, hcat(fill(i, length(yseq)), yseq))
    Plots.plot!(temp2[:, 1], temp2[:, 2], label = "", color = "gray", linewidth = 2)
end

Plots.plot!(Shape([-0.03, -0.018, -0.018, -0.03, -0.03].+0.0155,  [0.004, 0.004, 0.017, 0.017, 0.004].-0.0249),
            color = "white")
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155, [0.014, 0.014, 0.016, 0.016, 0.014].-0.0249), color = colScale[3])
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155, [0.011, 0.011, 0.013, 0.013, 0.011].-0.0249), color = colScale[4])
Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155, [0.008, 0.008, 0.01, 0.01, 0.008].-0.0249), color = colScale[5])

Plots.plot!(Shape([-0.029, -0.027, -0.027, -0.029, -0.029].+0.0155, [0.005, 0.005, 0.007, 0.007, 0.005].-0.0249), color = "white")


xseq = collect(-0.0285:0.0005:-0.0275).+0.0155
yseq = collect(0.0055:0.0005:0.0065).-0.0249

StatsPlots.scatter!(repeat(xseq, inner = 3), repeat(yseq, outer = 3),
                    markerstrokewidth = 0, markersize = 3, color = "black")

Plots.annotate!([(-0.026+0.0155, 0.015-0.0249, Plots.text("p = 0", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.012-0.0249, Plots.text("p = 1", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.009-0.0249, Plots.text("p = 2", :left, 24))])
Plots.annotate!([(-0.026+0.0155, 0.006-0.0249, Plots.text("q = 1", :left, 24))])

savefig(".\\Application 3\\Plots\\Order.pdf")

# Aggregating fitted values
λTotal = zeros(108)
for i = 1:36
    λTotal .+= results[mc[i], i].λ
end

Plots.plot(sum(dat, dims = 1)[1, :], label = "True", xlabel = "Time", ylabel = "Count")
Plots.plot!(λTotal, label = "λ")

# Single fitted values
i = 6

Plots.plot(DAT[i, 3], title = DAT[i, 1], label = "True", xlabel = "Time", ylabel = "Count")
xticks!(1:12:108, string.(2008:2016))
Plots.plot!(results[mc[i], i].λ, label = "λ")


# Check for how many districts PIT exceeds critical value
cival = quantile(Normal(), 1 - 0.05/20).*sqrt(108/10*0.9)/10.8
sum((i -> any(abs.(pit(results[mc[i], i], level = 0.95, nbins = 10)[2] .- 1) .> cival)).(1:36))

gr(size = (750, 500), linewidth = 1)
pit(results[mc[31], 31], nbins = 10, level = 0.95)
savefig(".\\Application 3\\Plots\\PIT.pdf")
