using Pkg, Plots, Shapefile, Luxor, LinearAlgebra

cd("C:\\Users\\stapperm\\sciebo\\Arbeit\\Projects\\Commuter")

# Ger = Shapefile.Table(".\\Data\\Shapefile\\LKGermany.shp")
# GerGeom = Shapefile.shapes(Ger)

# Basic projection function
function proj(focus, point::Vector{Float64})
    # Both [lat, long]
    θ = (90 - point[2])/180*π
    ϕ = point[1]/360*2*π

    x = sin(θ)*cos(ϕ)
    y = sin(θ)*sin(ϕ)
    z = cos(θ)

    θF = focus[2]/180*π
    ϕF = (90 + focus[1])/360*2*π

    # Rz = [cos(-ϕF) -sin(-ϕF) 0; sin(-ϕF) cos(-ϕF) 0; 0 0 1]
    # Rx = [1 0 0 ; 0 cos(θF) -sin(θF); 0 sin(θF) cos(θF)]
    R = [cos(-ϕF) -sin(-ϕF) 0; sin(θF)*sin(-ϕF) sin(θF)*cos(-ϕF) cos(θF)]
    R*[x, y, z]
end

function proj(focus, point::Array{Float64})
    if !(2 in size(point))
        error("Wrong dimensions")
    end

    transposed = false
    if size(point)[1] != 2
        transposed = true
        point = transpose(point)
    end

    # Both [lat, long]
    θ = (90 .- point[2, :])./180 .* π
    ϕ = point[1, :]./360 .* 2 .* π

    x = sin.(θ).*cos.(ϕ)
    y = sin.(θ).*sin.(ϕ)
    z = cos.(θ)

    θF = focus[2]/180*π
    ϕF = (90 + focus[1])/360*2*π

    # Rz = [cos(-ϕF) -sin(-ϕF) 0; sin(-ϕF) cos(-ϕF) 0; 0 0 1]
    # Rx = [1 0 0 ; 0 cos(θF) -sin(θF); 0 sin(θF) cos(θF)]
    R = [cos(-ϕF) -sin(-ϕF) 0; sin(θF)*sin(-ϕF) sin(θF)*cos(-ϕF) cos(θF)]
    res = R*vcat(x', y', z')
    if transposed
        res = convert(Array{Float64, 2}, res')
    end
    res
end

function proj(focus, poly::Shapefile.Polygon)
    xx = (x -> x.x).(poly.points)
    yy = (x -> x.y).(poly.points)

    newPointsRaw = proj(focus, hcat(xx, yy))
    newPoints = (i -> Shapefile.Point(newPointsRaw[i, 1], newPointsRaw[i, 2])).(1:length(xx))

    l = minimum(newPointsRaw[1, :])
    r = maximum(newPointsRaw[1, :])
    b = minimum(newPointsRaw[2, :])
    t = maximum(newPointsRaw[2, :])
    newMBR = Shapefile.Rect(l, b, r, t)

    Shapefile.Polygon(newMBR, poly.parts, newPoints)
end

# plot(GerGeom)
#
# gr(size = (500, 600))
# plot(proj.(fill([10, 50], length(GerGeom)), GerGeom), color = :transparent,
#      grid = false, xaxis = false, yaxis = false)
# #
# for i = collect(47:55)
#     temp = proj([10, 50], hcat(collect(5:0.1:15), fill(i, 101)))
#     plot!(temp[:, 1], temp[:, 2], label = "", color = "gray")
# end
#
# for i = collect(5:15)
#     temp = proj([10, 50], hcat(fill(i, 81), collect(47:0.1:55)))
#     plot!(temp[:, 1], temp[:, 2], label = "", color = "gray")
# end
# plot!(proj.(fill([10, 50], length(GerGeom)), GerGeom), color = :transparent)
