library(tscount)

singleRun = function(TT, a, b, initiate){
  b0 = 15*(1 - a - b)
  model = list(past_obs = 1, past_mean = 1)
  truepars = list(intercept = b0, past_obs = a, past_mean = b)
  y = as.numeric(tsglm.sim(TT, param = truepars, model = model)$ts)
  t1 = Sys.time()
  res = tsglm(ts = y, model = model, init.method = initiate)
  esttime = as.numeric(difftime(Sys.time(), t1, unit = "secs"))
  est = res$coef
  
  return(list(est = est, esttime = esttime))
}

set.seed(1)
singleRun(2000, 0.4, 0.3, "firstobs")
set.seed(1)
singleRun(2000, 0.4, 0.3, "iid")
set.seed(1)
singleRun(2000, 0.4, 0.3, "marginal")

combis = array(0, dim = c(36, 2))
counter = 1
for(i in seq(0.1, 0.8, by = 0.1)){
  for(j in seq(0.1, 0.8, by = 0.1)){
    if(i + j < 1){
      combis[counter, ] = c(i, j)
      counter = counter + 1
    }
  }
}

ESTS = array(0, dim = c(36, 3, 3, 3, 1000))
ESTTIMES = array(0, dim = c(36, 3, 3, 1000))

for(tt in 1:3){
  TTemp = c(50, 200, 1000)[tt]
  for(init in 1:3){
    initiateTemp = c("marginal", "firstobs", "iid")[init]
    
    for(cc in 1:36){
      for(i in 1:1000){
        set.seed(i*1000 + cc)
        temp = singleRun(TTemp, combis[cc, 1], combis[cc, 2], initiateTemp)
        ESTS[cc, tt, init, , i] = temp$est
        ESTTIMES[cc, tt, init, i] = temp$esttime
        print(c(tt, init, cc, i))
      }
    }
  }
}

mean(ESTTIMES[, 1, , ]) # 0.1071967
mean(ESTTIMES[, 2, , ]) # 0.1768961
mean(ESTTIMES[, 3, , ]) # 0.5470441
