using Pkg
# Adding CountTimeSeries package in Pkg REPL
# add https://github.com/ManuelStapper/CountTimeSeries.jl
using CountTimeSeries, Plots, StatsPlots, Random, Distributions, LaTeXStrings

gr(tickfontsize = 9, guidefontsize = 16, linewidth = 2)
cd("C:\\Users\\stapperm\\sciebo\\Arbeit\\Projects\\CountTimeSeries_Review")

function singleRun(T::Int64, α1::Float64, β1::Float64, initiate::String)
    β0 = 15*(1 - α1 - β1)
    model = Model(pastObs = 1, pastMean = 1)
    θ = θ2par([β0, α1, β1], model)
    y = simulate(T, model, θ)[1]
    setting = MLESettings(y, model, θ, optimizer = "NelderMead")
    times = zeros(3)

    t1 = time()
    est = fit(y, model, setting, initiate = initiate, printResults = false).θ
    esttime = time() - t1

    return est, esttime
end

combis = zeros(1000, 2)
counter = 1
for i = 0.1:0.1:0.8
    global counter
    for j = 0.1:0.1:0.8
        if i + j < 1
            combis[counter, :] = [i, j]
            counter += 1
        end
    end
end

combis = combis[1:counter-1, :]

ESTS = zeros(size(combis)[1], 3, 3, 3, 1000)
ESTTIMES = zeros(size(combis)[1], 3, 3, 1000)

for t = 1:3
    TTemp = [50, 200, 1000][t]
    for initi = 1:3
        initiTemp = ["marginal", "first", "intercept"][initi]
        for c = 1:size(combis)[1]
            for i = 1:1000
                Random.seed!(i*1000 + c)
                temp = singleRun(TTemp, combis[c, 1], combis[c, 2], initiTemp)
                ESTS[c, t, initi, :, i] = temp[1]
                ESTTIMES[c, t, initi, i] = temp[2]
                println([t, initi, c, i])
            end
        end
    end
end

mean(ESTTIMES[:, 1, :, :]) # 0.0032693243755234614
mean(ESTTIMES[:, 2, :, :]) # 0.007396305923108701
mean(ESTTIMES[:, 3, :, :]) # 0.02495582437736017

# Revise from here

# Compared to R:
0.1071967/0.0032693243755234614
0.1768961/0.007396305923108701
0.5470441/0.02495582437736017

scatter(0.2:0.1:0.9, (a -> sum(ESTTIMES[combis[:, 1] .+ combis[:, 2] .== a, 1, :, :])).(0.2:0.1:0.9))

scatter(0.1:0.1:0.8, (a -> sum(ESTTIMES[combis[:, 1] .== a, 1, :, :])).(0.1:0.1:0.8), ylims = (0, 150))
scatter(0.1:0.1:0.8, (a -> sum(ESTTIMES[combis[:, 2] .== a, 1, :, :])).(0.1:0.1:0.8), ylims = (0, 150))

sum(ESTTIMES[:, 2, 1, :])./sum(ESTTIMES[:, 2, 3, :])

# Comparision intitial as parameter

using Optim

function ll2(y, θ)
    T = length(y)

    if θ[2] + θ[3] > 1
        return -Inf
    end
    if any(θ .< 0)
        return -Inf
    end

    λ = zeros(T)

    λ[1] = θ[4]

    for t = 2:T
        λ[t] = θ[1] + θ[2]*y[t-1] + θ[3]*λ[t-1]
    end

    sum(logpdf.(Poisson.(λ), y))
end

function estFun(y, init)
    T = length(y)

    optRaw = optimize(vars -> -ll2(y, vars), init, NelderMead(), Optim.Options(iterations =  100000))

    return optRaw.minimizer
end


MarPar = zeros(2, 3, 36, 3, 1000)
MarParTime = zeros(2, 3, 36, 1000)

for t = 1:3
    TTemp = [50, 200, 1000][t]
    for c = 1:size(combis)[1]
        α1 = combis[c, 1]
        β1 = combis[c, 2]
        for i = 1:1000
            Random.seed!(i*1000 + c)
            temp = singleRun(TTemp, combis[c, 1], combis[c, 2], "marginal")
            MarPar[1, t, c, :, i] = temp[1]
            MarParTime[1, t, c, i] = temp[2]

            Random.seed!(i*1000 + c)
            β0 = 15*(1 - α1 - β1)
            model = Model(pastObs = 1, pastMean = 1)
            θ = θ2par([β0, α1, β1], model)
            y = simulate(TTemp, model, θ)[1]
            t1 = time()
            MarPar[2, t, c, :, i] = estFun(y, [β0, α1, β1, y[1]])[1:3]
            MarParTime[2, t, c, i] = time() - t1

            println([t, c, i])
        end
    end
end

# ESTS: combination, t, initial, parameter, run
# ESTSTIMES: combination, t, initial, run
# MarPar: initial, t, combination, parameter, run

# Combine both and average estimates
# MESTS: combination, t, initial, parameter
MESTS = zeros(36, 3, 4, 3)

MESTS[:, :, 1:3, :] = mean(ESTS, dims = 5)[:, :, :, :, 1]
for c = 1:36
    for t = 1:3
        MESTS[c, t, 4, :] = mean(MarPar[2, t, c, :, :], dims = 2)[:, 1]
    end
end

# MESTS: combination, t, initial, parameter

plot(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 1, 1, 2]) - a).(0.1:0.1:0.8),
     color = 1, label = "Marginal", xlabel = L"\alpha_1", ylabel = "Bias", legend = :bottomleft)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 1, 2, 2]) - a).(0.1:0.1:0.8),
     color = 2, label = "First")
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 1, 3, 2]) - a).(0.1:0.1:0.8),
     color = 3, label = "Intercept")
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 1, 4, 2]) - a).(0.1:0.1:0.8),
     color = 4, label = "Parameter")
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 2, 1, 2]) - a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 2, 2, 2]) - a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 2, 3, 2]) - a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 2, 4, 2]) - a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 3, 1, 2]) - a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 3, 2, 2]) - a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 3, 3, 2]) - a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 1] .== a, 3, 4, 2]) - a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dot)

savefig("Application 4\\Plots\\BiasAlpha.pdf")

plot(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 1, 1, 3]) - a).(0.1:0.1:0.8),
     color = 1, label = "Marginal", xlabel = L"\beta_1", ylabel = "Bias", legend = :bottomleft)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 1, 2, 3]) - a).(0.1:0.1:0.8),
     color = 2, label = "First")
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 1, 3, 3]) - a).(0.1:0.1:0.8),
     color = 3, label = "Intercept")
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 1, 4, 3]) - a).(0.1:0.1:0.8),
     color = 4, label = "Parameter")
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 2, 1, 3]) - a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 2, 2, 3]) - a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 2, 3, 3]) - a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 2, 4, 3]) - a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 3, 1, 3]) - a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 3, 2, 3]) - a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 3, 3, 3]) - a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> mean(MESTS[combis[:, 2] .== a, 3, 4, 3]) - a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dot)
savefig("Application 4\\Plots\\BiasBeta.pdf")

plot(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 1, 1, 2]) - a)/a).(0.1:0.1:0.8),
     color = 1, label = "Marginal", xlabel = L"\alpha_1", ylabel = "Relative Bias")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 1, 2, 2]) - a)/a).(0.1:0.1:0.8),
     color = 2, label = "First")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 1, 3, 2]) - a)/a).(0.1:0.1:0.8),
     color = 3, label = "Intercept")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 1, 4, 2]) - a)/a).(0.1:0.1:0.8),
     color = 4, label = "Parameter")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 2, 1, 2]) - a)/a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 2, 2, 2]) - a)/a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 2, 3, 2]) - a)/a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 2, 4, 2]) - a)/a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 3, 1, 2]) - a)/a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 3, 2, 2]) - a)/a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 3, 3, 2]) - a)/a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 1] .== a, 3, 4, 2]) - a)/a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dot)
savefig("Application 4\\Plots\\BiasAlphaRel.pdf")

plot(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 1, 1, 3]) - a)/a).(0.1:0.1:0.8),
     color = 1, label = "Marginal", xlabel = L"\beta_1", ylabel = "Relative Bias", legend = :bottomleft)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 1, 2, 3]) - a)/a).(0.1:0.1:0.8),
     color = 2, label = "First")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 1, 3, 3]) - a)/a).(0.1:0.1:0.8),
     color = 3, label = "Intercept")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 1, 4, 3]) - a)/a).(0.1:0.1:0.8),
     color = 4, label = "Parameter")
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 2, 1, 3]) - a)/a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 2, 2, 3]) - a)/a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 2, 3, 3]) - a)/a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 2, 4, 3]) - a)/a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dash)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 3, 1, 3]) - a)/a).(0.1:0.1:0.8),
     color = 1, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 3, 2, 3]) - a)/a).(0.1:0.1:0.8),
     color = 2, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 3, 3, 3]) - a)/a).(0.1:0.1:0.8),
     color = 3, label = "", linestyle = :dot)
plot!(0.1:0.1:0.8, (a -> (mean(MESTS[combis[:, 2] .== a, 3, 4, 3]) - a)/a).(0.1:0.1:0.8),
     color = 4, label = "", linestyle = :dot)
savefig("Application 4\\Plots\\BiasBetaRel.pdf")
